import contextlib
import os
import re
import tempfile
from typing import cast

from docutils import nodes


# $set_env.py: LINKLINT_SAVE_INTERMEDIATE - Save intermediate data in files for debugging.
SAVE_INTERMEDIATE = os.getenv("LINKLINT_SAVE_INTERMEDIATE") == "1"


def slug_for_test() -> str:  # pragma: debugging
    """Get the short name of the current test, possibly the param id."""
    test_name = os.getenv("PYTEST_CURRENT_TEST", "unknown")
    m = re.search(r"::(?P<test>[\w_]+)(?P<param>\[.+?\])?(?: \(\w+\))$", test_name)
    assert m is not None
    if param := m["param"]:
        return re.sub(r"[^\w]", "-", param.strip("[]"))
    return m["test"]


@contextlib.contextmanager
def in_tempdir():
    """Make a temp directory and cd into it, then come back and delete it."""
    with tempfile.TemporaryDirectory() as tmpdir:
        with contextlib.chdir(tmpdir):
            yield


def plural(n: int, thing: str = "", things: str = "") -> str:
    """Pluralize a word.

    If n is 1, return thing.  Otherwise return things, or thing+s.
    """
    if n == 1:
        noun = thing
    else:
        noun = things or (thing + "s")
    return f"{n} {noun}"


def node_line_number(node: nodes.Node) -> int:
    """Find a line number for a node, looking at parents if needed."""
    while node.line is None:
        node = cast(nodes.Node, node.parent)
    return node.line


def node_traceback(node: nodes.Node) -> str:  # pragma: debugging
    """Get a multi-line traceback for a node to pinpoint problems."""
    lines = []
    while node is not None:
        lineno = str(node.line) if node.line is not None else ""
        lines.append(f"{lineno:>4}: {node!r}")
        node = cast(nodes.Node, node.parent)
    return "\n".join(reversed(lines))



from dataclasses import dataclass
from typing import Iterable

from docutils import nodes
from sphinx import addnodes


@dataclass
class Region:
    # "module", "class", "function", etc.
    kind: str
    name: str
    # Line number where the region starts.
    start: int
    # Line number where the main part of the region ends: the lines in this
    # region and not in any nested subregion.
    end_main: int
    # Line number where the whole region ends: includes all nested subregions.
    end_total: int

    def __lt__(self, other):
        return (self.start, self.name) < (other.start, other.name)


class RegionFinder:
    def __init__(self) -> None:
        self.last_line = 0

    def find_regions(self, node: nodes.Node) -> Iterable[Region]:
        kind = None
        end_main = None

        match node:
            case nodes.section():
                section_names = set(node.get("names", []))
                for section_id in node.get("ids", []):
                    if section_id.startswith("module-") and section_id not in section_names:
                        kind = "module"
                        assert node.line is not None
                        name_starts = [(section_id[len("module-") :], node.line - 1)]
                        break

            case addnodes.desc():
                kind = node.get("objtype")
                name_starts = [
                    (kid.get("fullname"), kid.line)
                    for kid in node.children
                    if isinstance(kid, addnodes.desc_signature)
                ]

        last_line = getattr(node, "line", None)
        if last_line is not None:
            self.last_line = last_line + node.astext().count("\n")

        for child in node.children:
            for subregion in self.find_regions(child):
                if kind is not None and end_main is None:
                    # The main part of a region ends when the next region starts.
                    end_main = subregion.start - 1
                yield subregion

        if kind is not None:
            for name, start in name_starts:
                yield Region(
                    kind=kind,
                    name=name,
                    start=start,
                    end_main=end_main or self.last_line,
                    end_total=self.last_line,
                )


def find_regions(doctree: nodes.document) -> Iterable[Region]:
    """Find regions in the doctree."""
    finder = RegionFinder()
    return finder.find_regions(doctree)



"""Utilities for RST files."""

import re
import string
from pathlib import Path

from docutils import nodes
from sphinx.application import Sphinx

from linklint.dump import dump_doctree
from linklint.utils import SAVE_INTERMEDIATE, in_tempdir, slug_for_test


def run_sphinx(content: str, buildername: str, extensions: list[str]) -> nodes.document:
    # Create minimal conf.py
    Path("conf.py").write_text(f"extensions = {extensions!r}\n", encoding="utf-8")

    # Copy the RST file as index.rst
    Path("index.rst").write_text(content, encoding="utf-8")

    app = Sphinx(
        srcdir=".",
        confdir=".",
        outdir="_build",
        doctreedir="_build/.doctrees",
        buildername=buildername,
        freshenv=True,
        status=None,
        warning=None,
    )

    app.build()
    return app.env.get_doctree("index")


def parse_rst(content: str) -> nodes.document:
    """Parse RST content using Sphinx and return the doctree."""
    with in_tempdir():
        doctree = run_sphinx(content, buildername="dummy", extensions=[])
    fix_node_lines(doctree)
    if SAVE_INTERMEDIATE:
        save_test_doctree(doctree)
    return doctree


def save_test_doctree(doctree: nodes.document) -> None:
    dump_file = Path(f"tmp/doctree/{slug_for_test()}.txt")
    dump_file.parent.mkdir(parents=True, exist_ok=True)
    with dump_file.open("w", encoding="utf-8") as f:
        dump_doctree(doctree, f)


BLOCK_NODES = (nodes.paragraph,)


def fix_node_lines(doctree: nodes.document) -> None:
    """Fix line numbers for inline nodes within block elements.

    Doctree block nodes have correct line numbers, but their inline children
    inherit the block's start line.  Walk each block's descendants counting
    newlines in text nodes to compute each inline node's actual line.
    """
    for block in doctree.findall(lambda n: isinstance(n, BLOCK_NODES)):
        if not block.line:
            continue
        newline_count = 0
        for node in block.findall():
            if node is block:
                continue
            node.line = block.line + newline_count
            if isinstance(node, nodes.Text):
                # References are trimmed, so we can miss a trailing newline.
                # The parent might have better information.
                if isinstance(node.parent, (nodes.inline, nodes.reference)):
                    newline_count += node.parent.rawsource.count("\n")
                else:
                    newline_count += str(node).count("\n")


def is_header_line(line: str, text_line: str) -> bool:
    """Check if a line is a header underline/overline for `text_line`."""
    stripped = line.rstrip()
    return (
        len(stripped) >= len(text_line.rstrip())
        and len(set(stripped)) == 1
        and stripped[0] in string.punctuation
    )


def replace_rst_line(lines: list[str], line_num: int, new_line: str) -> None:
    """Replace a line in the content lines, adjusting header lengths if needed."""
    old_line = lines[line_num]
    lines[line_num] = new_line
    # Adjust adjacent lines if they are header lines.
    for adj_line_num in [line_num - 1, line_num + 1]:
        if adj_line_num in range(len(lines)):
            adj_line = lines[adj_line_num]
            if is_header_line(adj_line, old_line):
                line_end = adj_line[len(adj_line.rstrip()) :]
                lines[adj_line_num] = adj_line[0] * len(new_line.rstrip()) + line_end


def resub_in_rst_line(lines: list[str], line_num: int, pat: str, repl: str, count=0) -> bool:
    """Replace a substring in a line, adjusting header lengths if needed."""
    new_line = re.sub(pat, repl, lines[line_num], count=count)
    changed = new_line != lines[line_num]
    if changed:
        replace_rst_line(lines, line_num, new_line)
    return changed



import pytest

from linklint.regions import Region, find_regions
from linklint.rsthelp import parse_rst

from helpers import text_and_id


def region_test_case(*, rst: str, regions: list[Region], id: str = ""):
    """Helper to create pytest parameters for tests."""
    rst, id = text_and_id(text=rst, id=id)
    return pytest.param(rst, regions, id=id)


def region(
    kind: str,
    name: str,
    *,
    start: int,
    end: int = 0,
    end_main: int = 0,
    end_total: int = 0,
) -> Region:
    """Helper to create a Region with end_main and end_total defaulting to end."""
    assert end or (end_main and end_total), (
        "Either end or both end_main and end_total must be provided"
    )
    if not end_main:
        end_main = end
    if not end_total:
        end_total = end
    return Region(kind=kind, name=name, start=start, end_main=end_main, end_total=end_total)


TEST_CASES = [
    # The `contents::` directive makes unnumbered paragraphs.
    region_test_case(
        id="unnumbered",
        rst="""
            ======================
            Design and History FAQ
            ======================

            .. only:: html

               .. contents::

            Why does Python use indentation for grouping of statements?
            -----------------------------------------------------------

            Guido van Rossum believes that using indentation for grouping is extremely
            elegant and contributes a lot to the clarity of the average Python program.
            Most people learn to love this feature after a while.
            """,
        regions=[],
    ),
    region_test_case(
        id="mymodule",
        rst="""
            1My Module
            ==========

            .. module:: mymodule

            6This is the :mod:`mymodule` documentation.

            8Also check :mod:`othermodule` for related functionality.
            lorem ipsum quia dolor sit amet consectetur adipisci velit, sed quia non
            numquam eius modi tempora incidunt, ut labore et dolore magnam aliquam quaerat
            voluptatem. Ut enim ad minima veniam, quis nostrum exercitationem ullam
            12corporis suscipit laboriosam.

            14more about this module
            ------------------------

            17lorem ipsum quia dolor sit amet consectetur adipisci velit, sed
            quia non numquam eius modi tempora incidunt.

            20Other Section
            ===============

            23This section references :mod:`mymodule` which is fine.
            """,
        regions=[
            region("module", "mymodule", start=1, end=18),
        ],
    ),
    region_test_case(
        id="multiple-functions",
        rst=r"""
            The various :func:`exec\* <execl>` functions take a list of arguments for the new
            program loaded into the process.

            .. function:: execl(path, arg0, arg1, ...)
                          execle(path, arg0, arg1, ..., env)
                          execlp(file, arg0, arg1, ...)
                          execlpe(file, arg0, arg1, ..., env)

               This is a family of similar functions.
            """,
        regions=[
            region("function", "execl", start=4, end=9),
            region("function", "execle", start=4, end=9),
            region("function", "execlp", start=4, end=9),
            region("function", "execlpe", start=4, end=9),
        ],
    ),
    region_test_case(
        rst="lzma.rst",
        regions=[
            region("exception", "LZMAError", start=26, end=29),
            region("function", "open", start=35, end=68),
            region("method", "LZMAFile.peek", start=108, end=117),
            region("attribute", "LZMAFile.mode", start=119, end=123),
            region("attribute", "LZMAFile.name", start=125, end=130),
            region("class", "LZMAFile", start=71, end_main=107, end_total=141),
            region("method", "LZMACompressor.compress", start=209, end=215),
            region("method", "LZMACompressor.flush", start=217, end=222),
            region("class", "LZMACompressor", start=147, end_main=208, end_total=222),
            region("method", "LZMADecompressor.decompress", start=254, end=279),
            region("attribute", "LZMADecompressor.check", start=281, end=285),
            region("attribute", "LZMADecompressor.eof", start=287, end=289),
            region("attribute", "LZMADecompressor.unused_data", start=291, end=295),
            region("attribute", "LZMADecompressor.needs_input", start=297, end=302),
            region("class", "LZMADecompressor", start=225, end_main=253, end_total=302),
            region("function", "compress", start=304, end=310),
            region("function", "decompress", start=313, end=322),
            region("function", "is_check_supported", start=328, end=335),
            region("module", "lzma", start=1, end_main=25, end_total=346),
        ],
    ),
]


@pytest.mark.parametrize("rst, regions", TEST_CASES)
def test_regions(rst: str, regions: list[Region]) -> None:
    assert sorted(find_regions(parse_rst(rst))) == sorted(regions)



from difflib import Differ
from textwrap import dedent
from typing import Any

import pytest

from linklint.linklint import LintIssue, lint_content

from helpers import text_and_id


def diff_lines(text1: str, text2: str) -> str:
    """Return a diff of just the lines that differ between text1 and text2."""
    differ = Differ().compare(
        text1.splitlines(keepends=True),
        text2.splitlines(keepends=True),
    )
    min_diff = "".join(line for line in differ if line.startswith(("-", "+")))
    return min_diff


def lint_test_case(
    *,
    rst: str,
    issues: list[LintIssue],
    diff: str = "",
    id: str = "",
    marks: Any = None,
):
    """Helper to create pytest parameters for linting tests."""
    rst, id = text_and_id(text=rst, id=id)
    if diff.startswith("\n"):
        diff = diff[1:]
    return pytest.param(rst, issues, dedent(diff), id=id, marks=marks or [])


SELF_TEST_CASES = [
    # Check a self-link in the module description.
    lint_test_case(
        id="selflink",
        rst="""
            My Module
            =========

            .. module:: mymodule

            This is the :mod:`mymodule` documentation.

            Also check :mod:`othermodule` for related functionality.

            Other Section
            =============

            This section references :mod:`mymodule` which is fine.
            """,
        issues=[
            LintIssue(6, "self-link to :mod:`mymodule`", fixed=True),
        ],
        diff="""
            - This is the :mod:`mymodule` documentation.
            + This is the :mod:`!mymodule` documentation.
            """,
    ),
    # Is the line-munging correct?
    lint_test_case(
        id="second-section",
        rst="""
            Another
            =======

            Look at :mod:`mymodule` for more info.

            My Module
            =========

            .. module:: mymodule

            This is the :mod:`mymodule` documentation.

            Also check :mod:`othermodule` for related functionality.

            Other Section
            =============

            This section references :mod:`mymodule` which is fine.
            """,
        issues=[
            LintIssue(11, "self-link to :mod:`mymodule`", fixed=True),
        ],
        diff="""
            - This is the :mod:`mymodule` documentation.
            + This is the :mod:`!mymodule` documentation.
            """,
    ),
    # Check that `.. _module-foo` isn't confused for a module section.
    lint_test_case(
        id="module-target",
        rst="""
            .. _module-xyzzy:

            About Xyzzy
            -----------

            Xyzzy is a magical word that does nothing in particular.
            See :mod:`xyzzy` for more info.
            """,
        issues=[],
    ),
    # Check that headers get fixed too.
    lint_test_case(
        id="header-selflink",
        rst="""
            :mod:`mymodule` Module
            ======================

            .. module:: mymodule

            This is a great module!
            """,
        issues=[
            LintIssue(1, "self-link to :mod:`mymodule`", fixed=True),
        ],
        diff="""
            - :mod:`mymodule` Module
            + :mod:`!mymodule` Module
            - ======================
            + =======================
            """,
    ),
    # Check that overline headers get fixed too.
    lint_test_case(
        id="overline",
        rst="""
            ***********************
            This is :mod:`mymodule`
            ***********************

            .. module:: mymodule

            This is a great module!

            .. deprecated:: 3.8

                Maybe it's not so great after all.
            """,
        issues=[
            LintIssue(2, "self-link to :mod:`mymodule`", fixed=True),
        ],
        diff="""
            - ***********************
            + ************************
            - This is :mod:`mymodule`
            + This is :mod:`!mymodule`
            - ***********************
            + ************************
            """,
    ),
    # Self-link on a continuation line inside a list item.
    lint_test_case(
        id="continuation-line",
        rst="""
            My Module
            =========

            .. module:: mymodule

            - First item talks about something.
              And continues with :mod:`mymodule` here.
            - Second item.
            """,
        issues=[
            LintIssue(7, "self-link to :mod:`mymodule`", fixed=True),
        ],
        diff="""
            -   And continues with :mod:`mymodule` here.
            +   And continues with :mod:`!mymodule` here.
            """,
    ),
    # Don't get confused about sub-modules.
    lint_test_case(
        id="submodule",
        rst="""
            :mod:`dbm` --- Interfaces to Unix "databases"
            =============================================

            .. module:: dbm
            :synopsis: Interfaces to various Unix "database" formats.

            **Source code:** :source:`Lib/dbm/__init__.py`

            --------------

            :mod:`dbm` is a generic interface to variants of the DBM database:

            * :mod:`dbm.sqlite3`
            * :mod:`dbm.gnu`
            * :mod:`dbm.ndbm`

            If none of these modules are installed, the
            slow-but-simple implementation in module :mod:`dbm.dumb` will be used. There
            is a `third party interface <https://www.jcea.es/programacion/pybsddb.htm>`_ to
            the Oracle Berkeley DB.

            .. note::
               None of the underlying modules will automatically shrink the disk space used by
               the database file. However, :mod:`dbm.sqlite3`, :mod:`dbm.gnu` and :mod:`dbm.dumb`
               provide a :meth:`!reorganize` method that can be used for this purpose.
            """,
        issues=[
            LintIssue(line=1, message="self-link to :mod:`dbm`", fixed=True),
            LintIssue(line=11, message="self-link to :mod:`dbm`", fixed=True),
        ],
        diff="""
            - :mod:`dbm` --- Interfaces to Unix "databases"
            + :mod:`!dbm` --- Interfaces to Unix "databases"
            - =============================================
            + ==============================================
            - :mod:`dbm` is a generic interface to variants of the DBM database:
            + :mod:`!dbm` is a generic interface to variants of the DBM database:
            """,
    ),
    # Fix ~ references if needed.
    lint_test_case(
        id="tilde",
        rst="""
            :mod:`email.encoders`: Encoders
            ----------------------------------

            .. module:: email.encoders

            This module is deprecated in Python 3.  The functions provided here
            should not be called explicitly since the :class:`~email.mime.text.MIMEText`
            class sets the content type and CTE header using the *_subtype* and *_charset*
            values passed during the instantiation of that class.

            The :mod:`email` package provides some convenient encoders in its
            :mod:`~email.encoders` module.  These encoders are actually used by the
            :class:`~email.mime.audio.MIMEAudio` and :class:`~email.mime.image.MIMEImage`
            class constructors to provide default encodings.
            """,
        issues=[
            LintIssue(line=1, message="self-link to :mod:`email.encoders`", fixed=True),
            LintIssue(line=12, message="self-link to :mod:`email.encoders`", fixed=True),
        ],
        diff="""
            - :mod:`email.encoders`: Encoders
            + :mod:`!email.encoders`: Encoders
            - ----------------------------------
            + --------------------------------
            - :mod:`~email.encoders` module.  These encoders are actually used by the
            + :mod:`!email.encoders` module.  These encoders are actually used by the
            """,
    ),
    # Fix dotted references
    lint_test_case(
        id="dotted",
        rst="""
            :mod:`!html.parser` --- Simple HTML and XHTML parser
            ====================================================

            .. module:: html.parser
               :synopsis: A simple parser that can handle HTML and XHTML.

            --------------

            This module defines a class :class:`HTMLParser` which serves as the basis for
            parsing text files formatted in HTML (HyperText Mark-up Language) and XHTML.

            .. class:: HTMLParser(*, convert_charrefs=True, scripting=False)

               Create a parser instance able to parse invalid markup.

               An :class:`.HTMLParser` instance is fed HTML data and calls handler methods
               when start tags, end tags, text, comments, and other markup elements are
               encountered.  The user should subclass :class:`.HTMLParser` and override its
               methods to implement the desired behavior.
            """,
        issues=[
            LintIssue(line=16, message="self-link to :class:`HTMLParser`", fixed=True),
            LintIssue(line=18, message="self-link to :class:`HTMLParser`", fixed=True),
        ],
        diff="""
            -    An :class:`.HTMLParser` instance is fed HTML data and calls handler methods
            +    An :class:`!HTMLParser` instance is fed HTML data and calls handler methods
            -    encountered.  The user should subclass :class:`.HTMLParser` and override its
            +    encountered.  The user should subclass :class:`!HTMLParser` and override its
            """,
    ),
    # Class self-linking.
    lint_test_case(
        id="selflink-class",
        rst="""
            Queue
            =====

            .. class:: Queue(maxsize=0)

               A first in, first out (FIFO) queue.

               .. method:: shutdown(immediate=False)

                  Put a :class:`Queue` instance into a shutdown mode.
            """,
        issues=[
            LintIssue(line=10, message="self-link to :class:`Queue`", fixed=True),
        ],
        diff="""
            -       Put a :class:`Queue` instance into a shutdown mode.
            +       Put a :class:`!Queue` instance into a shutdown mode.
            """,
    ),
    # Some implicit references have no line number?
    # Optional[Anchor] makes a reference to Anchor with no line number and
    # we can't fix it anyway.
    lint_test_case(
        id="implicit-ref",
        rst="""
            :mod:`!importlib.resources` -- Package resource reading, opening and access
            ---------------------------------------------------------------------------

            .. module:: importlib.resources

            .. class:: Anchor

                Represents an anchor for resources, either a :class:`module object
                <types.ModuleType>` or a module name as a string. Defined as
                ``Union[str, ModuleType]``.

            .. function:: files(anchor: Optional[Anchor] = None)

                *anchor* is an optional :class:`Anchor`. If the anchor is a
                package, resources are resolved from that package. If a module,
                resources are resolved adjacent to that module (in the same package
                or the package root). If the anchor is omitted, the caller's module
                is used.
            """,
        issues=[],
    ),
    # Some directives had the wrong line number.
    lint_test_case(
        id="note",
        # https://github.com/sphinx-doc/sphinx/pull/14309 fixes the line numbers
        # for versionchanged directives, so our line numbers are off until that
        # is merged.
        marks=pytest.mark.xfail,
        rst="""
            ZipFile objects
            ---------------

            .. class:: ZipFile(file, mode='r', compression=ZIP_STORED, allowZip64=True)

               Open a ZIP file, where *file* can be a path to a file (a string), a
               file-like object or a :term:`path-like object`.

               .. versionchanged:: 3.2
                  Added the ability to use :class:`ZipFile` as a context manager.
                  Also did many other good things.

               .. note::
                  Added the ability to use :class:`ZipFile` as a context manager.
                  Also did many other good things.
            """,
        issues=[
            LintIssue(line=10, message="self-link to :class:`ZipFile`", fixed=True),
            LintIssue(line=14, message="self-link to :class:`ZipFile`", fixed=True),
        ],
        diff="""
            -       Added the ability to use :class:`ZipFile` as a context manager.
            +       Added the ability to use :class:`!ZipFile` as a context manager.
            -       Added the ability to use :class:`ZipFile` as a context manager.
            +       Added the ability to use :class:`!ZipFile` as a context manager.
            """,
    ),
    # Sometimes there are two to fix in one line.
    lint_test_case(
        id="two-in-one-line",
        rst="""
            ZipFile objects
            ---------------

            .. class:: ZipFile(file, mode='r', compression=ZIP_STORED, allowZip64=True)

               This is :class:`ZipFile` and also :class:`ZipFile` again.
               It's great.
            """,
        issues=[
            LintIssue(line=6, message="self-link to :class:`ZipFile`", fixed=True),
            LintIssue(line=6, message="self-link to :class:`ZipFile`", fixed=True),
        ],
        diff="""
            -    This is :class:`ZipFile` and also :class:`ZipFile` again.
            +    This is :class:`!ZipFile` and also :class:`!ZipFile` again.
            """,
    ),
    # References can be in a section, but still be forward references, so
    # they aren't linking to the section you are already reading.
    lint_test_case(
        id="case-sensitive",
        rst="""
            :mod:`!uuid` --- UUID objects according to :rfc:`9562`
            ======================================================

            .. module:: uuid

            This module provides immutable :class:`UUID` objects (the :class:`UUID` class)
            and :ref:`functions <uuid-factory-functions>` for generating UUIDs corresponding
            to a specific UUID version as specified in :rfc:`9562` (which supersedes :rfc:`4122`),
            for example, :func:`uuid1` for UUID version 1, :func:`uuid3` for UUID version 3, and so on.
            Note that UUID version 2 is deliberately omitted as it is outside the scope of the RFC.

            .. class:: UUID(hex=None, bytes=None, bytes_le=None, fields=None, int=None, version=None, *, is_safe=SafeUUID.unknown)

               Create a UUID from either a string of 32 hexadecimal digits, a string of 16
            """,
        issues=[],
    ),
    # versionchanged directives needed fixing to get the right line numbers,
    # and the inline case needed extra fixing.
    lint_test_case(
        id="inline-versionchanged",
        # https://github.com/sphinx-doc/sphinx/pull/14309 fixes the line numbers
        # for versionchanged directives, so our line numbers are off until that
        # is merged.
        marks=pytest.mark.xfail,
        rst="""
            :mod:`!collections` --- Container datatypes
            ===========================================

            .. module:: collections

            :class:`Counter` objects
            ------------------------

            A counter tool is provided to support convenient and rapid tallies.

            .. class:: Counter([iterable-or-mapping])

                A :class:`!Counter` is a :class:`dict` subclass for counting :term:`hashable` objects.

                .. versionchanged:: 3.7 As a :class:`dict` subclass, :class:`Counter`
                   inherited the capability to remember insertion order.

                Counter objects support additional methods beyond those available for all dictionaries.

            """,
        issues=[
            LintIssue(line=15, message="self-link to :class:`Counter`", fixed=True),
        ],
        diff="""
            -     .. versionchanged:: 3.7 As a :class:`dict` subclass, :class:`Counter`
            +     .. versionchanged:: 3.7 As a :class:`dict` subclass, :class:`!Counter`
            """,
    ),
    # A newline at the end of a link is trimmed, so our line count was off.
    lint_test_case(
        id="newline-in-link",
        rst="""
            .. function:: fwalk(top='.', topdown=True, onerror=None, *, follow_symlinks=False, dir_fd=None)

               This function always supports :ref:`paths relative to directory descriptors
               <dir_fd>` and :ref:`not following symlinks <follow_symlinks>`.  Note however
               that, unlike other functions, the :func:`fwalk` default value for
               *follow_symlinks* is ``False``.

               This function always supports :ref:`paths relative to directory descriptors
               <dir_fd>`.  Note however that the :func:`fwalk` default value for
               *follow_symlinks* is ``False``.

            .. function:: expm1(x)

               Return *e* raised to the power *x*, minus 1.  Here *e* is the base of natural
               logarithms.  For small floats *x*, the subtraction in ``exp(x) - 1``
               can result in a `significant loss of precision
               <https://en.wikipedia.org/wiki/Loss_of_significance>`_; the :func:`expm1`
               function provides a way to compute this quantity to full precision.

            """,
        issues=[
            LintIssue(line=5, message="self-link to :func:`fwalk`", fixed=True),
            LintIssue(line=9, message="self-link to :func:`fwalk`", fixed=True),
            LintIssue(line=17, message="self-link to :func:`expm1`", fixed=True),
        ],
        diff="""
            -    that, unlike other functions, the :func:`fwalk` default value for
            +    that, unlike other functions, the :func:`!fwalk` default value for
            -    <dir_fd>`.  Note however that the :func:`fwalk` default value for
            +    <dir_fd>`.  Note however that the :func:`!fwalk` default value for
            -    <https://en.wikipedia.org/wiki/Loss_of_significance>`_; the :func:`expm1`
            +    <https://en.wikipedia.org/wiki/Loss_of_significance>`_; the :func:`!expm1`
            """,
    ),
    lint_test_case(
        id="stars",
        rst=r"""
            Process Management
            ------------------

            These functions may be used to create and manage processes.

            The various :func:`exec\* <execl>` functions take a list of arguments for the new
            program loaded into the process.

            .. function:: execl(path, arg0, arg1, ...)
                          execle(path, arg0, arg1, ..., env)
                          execlp(file, arg0, arg1, ...)
                          execlpe(file, arg0, arg1, ..., env)
                          execv(path, args)
                          execve(path, args, env)
                          execvp(file, args)
                          execvpe(file, args, env)

               The current process is replaced immediately. Open file objects and
               descriptors are not flushed, so if there may be data buffered
               on these open files, you should flush them using
               :func:`sys.stdout.flush` or :func:`os.fsync` before calling an
               :func:`exec\* <execl>` function.

               The "l" and "v" variants of the :func:`exec\* <execl>` functions differ in how
               command-line arguments are passed.  The "l" variants are perhaps the easiest
               to work with if the number of parameters is fixed when the code is written; the
               individual parameters simply become additional parameters to the :func:`!execl\*`
               functions.  The "v" variants are good when the number of parameters is
               variable, with the arguments being passed in a list or tuple as the *args*
               parameter.  In either case, the arguments to the child process should start with
               the name of the command being run, but this is not enforced.

               The variants which include a "p" near the end (:func:`execlp`,
               :func:`execlpe`, :func:`execvp`, and :func:`execvpe`) will use the
               :envvar:`PATH` environment variable to locate the program *file*.
            """,
        issues=[
            LintIssue(line=22, message="self-link to :func:`execl`", fixed=True),
            LintIssue(line=24, message="self-link to :func:`execl`", fixed=True),
            LintIssue(line=33, message="self-link to :func:`execlp`", fixed=True),
            LintIssue(line=34, message="self-link to :func:`execlpe`", fixed=True),
            LintIssue(line=34, message="self-link to :func:`execvp`", fixed=True),
            LintIssue(line=34, message="self-link to :func:`execvpe`", fixed=True),
        ],
        diff=r"""
            -    :func:`exec\* <execl>` function.
            +    :func:`!exec\*` function.
            -    The "l" and "v" variants of the :func:`exec\* <execl>` functions differ in how
            +    The "l" and "v" variants of the :func:`!exec\*` functions differ in how
            -    The variants which include a "p" near the end (:func:`execlp`,
            +    The variants which include a "p" near the end (:func:`!execlp`,
            -    :func:`execlpe`, :func:`execvp`, and :func:`execvpe`) will use the
            +    :func:`!execlpe`, :func:`!execvp`, and :func:`!execvpe`) will use the
            """,
    ),
    # A class mention in a method signature creates a reference with no line number, which
    # used to cause an assertion. But also, there's no markup for us to fix,
    # so it creates a LintIssue with fixed=False.
    lint_test_case(
        id="class-in-signature",
        rst="""
            Snapshot
            ^^^^^^^^

            .. class:: Snapshot

               Snapshot of traces of memory blocks allocated by Python.

               The :func:`take_snapshot` function creates a snapshot instance.

               .. method:: compare_to(old_snapshot: Snapshot, key_type: str, cumulative: bool=False)

                  Compute the differences with an old snapshot. Get statistics as a sorted
                  list of :class:`StatisticDiff` instances grouped by *key_type*.
            """,
        issues=[
            LintIssue(line=10, message="self-link to :class:`Snapshot`", fixed=False),
        ],
    ),
    lint_test_case(
        rst="classhier.rst",
        issues=[
            LintIssue(line=6, message="self-link to :class:`ClassA`", fixed=True),
            LintIssue(line=10, message="self-link to :class:`ClassA`", fixed=True),
            LintIssue(line=10, message="self-link to :meth:`methoda1`", fixed=True),
            LintIssue(line=14, message="self-link to :class:`ClassA`", fixed=True),
            LintIssue(line=18, message="self-link to :meth:`methodb1`", fixed=True),
            LintIssue(line=19, message="self-link to :class:`ClassA`", fixed=True),
            LintIssue(line=24, message="self-link to :class:`ClassA`", fixed=True),
            LintIssue(line=28, message="self-link to :meth:`methoda2`", fixed=True),
        ],
        diff=r"""
            -     An outer class, called :class:`ClassA`.
            +     An outer class, called :class:`!ClassA`.
            -         A method of :class:`ClassA`, called :meth:`methoda1`.
            +         A method of :class:`!ClassA`, called :meth:`!methoda1`.
            -         A nested class. This is inside :class:`ClassA`.
            +         A nested class. This is inside :class:`!ClassA`.
            -             A method of :class:`ClassB`, called :meth:`methodb1`.
            +             A method of :class:`ClassB`, called :meth:`!methodb1`.
            -             We're still inside :class:`ClassA`.
            +             We're still inside :class:`!ClassA`.
            -         We're still inside :class:`ClassA`.
            +         We're still inside :class:`!ClassA`.
            -     A method of :class:`ClassA`, called :meth:`methoda2`.
            +     A method of :class:`ClassA`, called :meth:`!methoda2`.
        """,
    ),
    lint_test_case(
        id="method-no-class",
        rst="""
            Memory-mapped file objects behave like both :class:`bytearray` and like
            :term:`file objects <file object>`.  You can use mmap objects in most places
            where :class:`bytearray` are expected; for example, you can use the :mod:`re`
            module to search through a memory-mapped file.  You can also change a single
            byte by doing ``obj[index] = 97``, or change a subsequence by assigning to a
            slice: ``obj[i1:i2] = b'...'``.  You can also read and write data starting at
            the current file position, and :meth:`seek` through the file to different positions.

            The easiest way to create a text stream is with :meth:`open`, optionally
            specifying an encoding.
            """,
        issues=[],
    ),
]


@pytest.mark.parametrize("rst, issues, diff", SELF_TEST_CASES)
def test_fix_self_link(rst: str, issues: list[LintIssue], diff: str) -> None:
    result = lint_content(rst, fix=True, checks={"self"})
    assert result.issues == issues
    assert diff_lines(rst, result.content) == diff


@pytest.mark.parametrize("rst, issues, diff", SELF_TEST_CASES)
def test_nofix_self_link(rst: str, issues: list[LintIssue], diff: str) -> None:
    result = lint_content(rst, fix=False, checks={"self"})
    issues = [LintIssue(issue.line, issue.message, fixed=False) for issue in issues]
    assert result.issues == issues
    assert rst == result.content


PARADUP_TEST_CASES = [
    lint_test_case(
        id="class-in-signature",
        rst="""
            LogRecord Objects
            -----------------

            :class:`LogRecord` instances are created automatically by the :class:`Logger`
            every time something is logged, and can be created manually via
            :func:`makeLogRecord` (for example, from a pickled event received over the
            wire).

            .. class:: LogRecord(name, level, pathname, lineno, msg, args, exc_info, func=None, sinfo=None)

               Contains all the information pertinent to the event being logged.

               :param args: Variable data to merge into the *msg* argument
                  to obtain the event description.
               :type args: tuple | dict[str, typing.Any]

               :param exc_info: An exception tuple with the current exception information,
                  as returned by :func:`sys.exc_info`,
                  or ``None`` if no exception information is available.
               :type exc_info: tuple[type[BaseException], BaseException, types.TracebackType] | None

               :param func: The name of the function or method
                  from which the logging call was invoked.
               :type func: str | None
            """,
        issues=[
            # Strange: line 2 is the underline of the header, the :param: is miles away.
            LintIssue(line=2, message="duplicate :class:`BaseException` in paragraph", fixed=False)
        ],
    ),
    lint_test_case(
        id="paragraphs",
        rst="""
            1 Multiple links to :class:`Something` in the same paragraph about
            :class:`Something` are not a problem, but they do create multiple
            LintIssues on different lines about :class:`Something`.

            5 If :class:`!Something` is referenced
            but not linked, then :class:`Something` could
            be the first link to :class:`Something` and the other links
            to :class:`Something` are duplicates.
            """,
        issues=[
            LintIssue(line=2, message="duplicate :class:`Something` in paragraph", fixed=False),
            LintIssue(line=3, message="duplicate :class:`Something` in paragraph", fixed=False),
            LintIssue(line=7, message="duplicate :class:`Something` in paragraph", fixed=False),
            LintIssue(line=8, message="duplicate :class:`Something` in paragraph", fixed=False),
        ],
    ),
    lint_test_case(
        id="lists",
        rst="""
            Some lists:

            - 3 Multiple links to :class:`Something` in the same paragraph about
              :class:`Something` are not a problem, but they do create multiple
              LintIssues on different lines about :class:`Something`.

            - 7 If :class:`!Something` is referenced
              but not linked, then :class:`Something` could
              be the first link to :class:`Something` and the other links
              to :class:`Something` are duplicates.
            """,
        issues=[
            LintIssue(line=4, message="duplicate :class:`Something` in paragraph", fixed=False),
            LintIssue(line=5, message="duplicate :class:`Something` in paragraph", fixed=False),
            LintIssue(line=9, message="duplicate :class:`Something` in paragraph", fixed=False),
            LintIssue(line=10, message="duplicate :class:`Something` in paragraph", fixed=False),
        ],
    ),
]


@pytest.mark.parametrize("rst, issues, diff", PARADUP_TEST_CASES)
def test_paradup(rst: str, issues: list[LintIssue], diff: str) -> None:
    assert diff == ""
    result = lint_content(rst, fix=False, checks={"paradup"})
    assert result.issues == issues



import argparse
import sys
from pathlib import Path

from linklint.linklint import CHECKS, LintIssue, lint_content
from linklint.utils import plural


def lint_file(filepath: str, fix: bool, checks: set[str]) -> list[LintIssue]:
    """Lint a single RST file.

    Returns a list of LintIssue objects.
    """
    # print(filepath)
    path = Path(filepath)
    content = path.read_text(encoding="utf-8")
    result = lint_content(content, fix, checks)
    if fix and result.fixed:
        path.write_text(result.content, encoding="utf-8")
    return result.issues


def linklint(argv: list[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--check",
        help=f"comma-separated checks to run ({', '.join(CHECKS)}, all)",
        default="all",
    )
    parser.add_argument("--fix", help="Fix the issues in place", action="store_true")
    parser.add_argument("files", nargs="+", help="RST files to lint")
    args = parser.parse_args(argv)

    checks = set(args.check.split(","))
    unknown = checks - set(CHECKS.keys()) - {"all"}
    if unknown:
        print(f"Unknown checks: {', '.join(unknown)}", file=sys.stderr)
        return 2
    if "all" in checks:
        checks = set(CHECKS.keys())

    issues = 0
    fixed = 0
    for filepath in args.files:
        # This runs Sphinx on each file separately, which seems slow, but is
        # faster than running it once on all the files.
        for issue in lint_file(filepath, args.fix, checks):
            fixed_suffix = ""
            if issue.fixed:
                fixed_suffix = " (fixed)"
                fixed += 1
            print(f"{filepath}:{issue.line}: {issue.message}{fixed_suffix}")
            issues += 1

    summary = f"Checked {plural(len(args.files), 'file')}, found {plural(issues, 'issue')}"
    if args.fix:
        summary += f", fixed {fixed}"
    print(f"{summary}.")

    return issues > 0


def main():
    sys.exit(linklint(sys.argv[1:]))



import sys
from typing import TextIO

from docutils import nodes


INTERESTING_KEYS = [
    "ids",
    "names",
    "reftype",
    "reftarget",
    "refuri",
    "domain",
    "objtype",
    "desctype",
    "class",
    "fullname",
    "module",
]

# INTERESTING_ATTRS = [
#     "rawsource",
# ]


def dump_doctree(node: nodes.Node, fp: TextIO, indent: int = 0) -> None:  # pragma: debugging
    """Print a nicely formatted tree of a docutils doctree."""
    prefix = "    " * indent
    if isinstance(node, nodes.Text):
        text = node.astext()
        print(f"{prefix}Text: {text!r}", file=fp)
    else:
        tag = node.__class__.__name__
        attrs = []
        for key in INTERESTING_KEYS:
            if val := node.get(key):  # type: ignore
                attrs.append(f"{key}={val!r}")
        # for attr in INTERESTING_ATTRS:
        #     if val := getattr(node, attr, None):
        #         attrs.append(f".{attr}={val!r}")
        attr_str = f" {{{', '.join(attrs)}}}" if attrs else ""
        if tag == "reference":
            print(vars(node), file=fp)
        line = node.line
        line_str = f" @{line}" if line else ""
        print(f"{prefix}{tag}{attr_str}{line_str}", file=fp)
        # print(f"{prefix}{node.attributes}", file=fp)
        for child in node.children:
            dump_doctree(child, fp, indent + 1)


if __name__ == "__main__":
    from linklint.linklint import parse_rst

    with open(sys.argv[1]) as f:
        dump_doctree(parse_rst(f.read()), sys.stdout)



from linklint.rsthelp import is_header_line

import pytest


@pytest.mark.parametrize(
    "hline, is_header",
    [
        ("======", True),
        ("------", True),
        ("======  ", True),  # Trailing space is ok
        ("======\n", True),  # Newline is ok
        ("  ======  ", False),  # Can't be indented
        ("====", False),  # Too short
        ("===========", True),  # Longer is ok
        ("=====-", False),  # Mixed chars
        ("123456", False),  # Not punctuation
    ],
)
def test_is_header_line(hline: str, is_header: bool) -> None:
    assert is_header_line(hline, "Header") == is_header



from typing import cast

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.util import logging
from sphinx.util.typing import ExtensionMetadata

import linklint
from linklint.linklint import find_duplicate_refs, find_self_refs

logger = logging.getLogger(__name__)


def init_data(app: Sphinx) -> None:
    app.env.linklint_counts: dict[str, int] = {}


def process_reference_nodes(app, doctree):
    """Find references that shouldn't be links, and un-reference them."""
    # Change <reference><literal></reference> to just <literal>.

    count = 0
    for finder in (find_self_refs, find_duplicate_refs):
        for ref in finder(doctree):
            cast(nodes.Element, ref.parent).replace(ref, ref.children[0])
            count += 1
    app.env.linklint_counts[app.env.docname] = count


def merge_data(
    app: Sphinx,
    env: BuildEnvironment,
    docnames: list[str],
    other: BuildEnvironment,
) -> None:
    # Env's start as copies of other envs, so we have to be careful to only
    # merge data for the docnames that are actually being merged.
    for docname in docnames:
        if docname in other.linklint_counts:  # type: ignore
            env.linklint_counts[docname] = other.linklint_counts[docname]  # type: ignore


def display_results(app: Sphinx, exception: Exception | None) -> None:
    if not exception:
        total = sum(app.env.linklint_counts.values())  # type: ignore
        logger.info(f"Linklint: unlinked {total} references")


def setup(app: Sphinx) -> ExtensionMetadata:
    app.connect("builder-inited", init_data)
    app.connect("doctree-read", process_reference_nodes)
    app.connect("env-merge-info", merge_data)
    app.connect("build-finished", display_results)

    return {
        "version": linklint.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
Tests using .rst files in tests/data.

Sphinx renders them to HTML, then we compare a summary of the HTML to an
expected summary in a corresponding .html file.

"""

import shutil
from pathlib import Path

import pytest

from linklint.rsthelp import run_sphinx, save_test_doctree
from linklint.utils import SAVE_INTERMEDIATE, in_tempdir

from summarize_html import summarize_html_file


PROJECT = Path(__file__).parent.parent


@pytest.mark.parametrize(
    "rst_file",
    [r.name for r in Path("tests/data").glob("*.rst")],
)
def test_summarize_html(rst_file: str) -> None:
    rst = (PROJECT / "tests/data" / rst_file).read_text(encoding="utf-8")
    root = rst_file.removesuffix(".rst")
    with in_tempdir():
        doctree = run_sphinx(rst, buildername="html", extensions=["linklint.ext"])
        summary = summarize_html_file("_build/index.html")
        if SAVE_INTERMEDIATE:
            # In case of needing to see what happened, copy the HTML etc to tmp.
            shutil.copytree("_build/_static", PROJECT / "tmp/html/_static", dirs_exist_ok=True)
            shutil.copyfile("_build/index.html", PROJECT / f"tmp/html/{root}.html")
            (PROJECT / f"tmp/html/{root}_summary.html").write_text(summary, encoding="utf-8")

            # Also run without the extension to understand Sphinx native behavior.
            run_sphinx(rst, buildername="html", extensions=[])
            shutil.copyfile("_build/index.html", PROJECT / f"tmp/html/{root}_nofix.html")
            nofix_summary = summarize_html_file("_build/index.html")
            (PROJECT / f"tmp/html/{root}_summary_nofix.html").write_text(
                nofix_summary, encoding="utf-8"
            )
            save_test_doctree(doctree)

    assert 'class="self-link"' not in summary, f"Self-links found in {root}.html"
    summary_file = PROJECT / f"tests/data/{root}_summary.html"
    if summary_file.exists():
        expected = summary_file.read_text(encoding="utf-8")
    else:
        expected = f"Summary {summary_file} doesn't exist"  # pragma: only failure

    if summary != expected:  # pragma: only failure
        if SAVE_INTERMEDIATE:
            print(f"Full HTML is at tmp/html/{root}.html")
            print(f"Summary is at tmp/html/{root}_summary.html")
            print("if the full HTML is correct:")
            print(f"  $ cp tmp/html/{root}_summary.html tests/data/{root}_summary.html")
        else:
            print("To see full HTML, set LINKLINT_SAVE_INTERMEDIATE=1 and re-run the test.")

    assert summary == expected



"""Linter to find link problems in RST files."""

import collections
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Iterable

from docutils import nodes
from sphinx import addnodes

from linklint.regions import Region, find_regions
from linklint.rsthelp import parse_rst, resub_in_rst_line
from linklint.utils import node_line_number, node_traceback


class Resolver:
    # Build a map from reference roles to object types.

    # Dummy lambda's so that the object_types dict can be identical to the
    # code in sphinx/sphinx/domains/python/__init__.py.
    ObjType = lambda _, *refs: refs  # noqa: E731
    _ = lambda s: 0  # noqa: E731

    # object_types map from sphinx
    # sphinx/sphinx/domains/python/__init__.py:725
    object_types = {
        "function": ObjType(_("function"), "func", "obj"),
        "data": ObjType(_("data"), "data", "obj"),
        "class": ObjType(_("class"), "class", "exc", "obj"),
        "exception": ObjType(_("exception"), "exc", "class", "obj"),
        "method": ObjType(_("method"), "meth", "obj"),
        "classmethod": ObjType(_("class method"), "meth", "obj"),
        "staticmethod": ObjType(_("static method"), "meth", "obj"),
        "attribute": ObjType(_("attribute"), "attr", "obj"),
        "property": ObjType(_("property"), "attr", "_prop", "obj"),
        "type": ObjType(_("type alias"), "type", "class", "obj"),
        "module": ObjType(_("module"), "mod", "obj"),
    }

    reftype_to_objtype = collections.defaultdict(list)
    for objtype, names in object_types.items():
        for name in names:
            reftype_to_objtype[name].append(objtype)

    def __init__(self, doctree: nodes.document) -> None:
        self.region_map = {(r.kind, r.name): r for r in find_regions(doctree)}

    def find_region(self, reftype: str, target: str) -> Region | None:
        for objtype in self.reftype_to_objtype[reftype]:
            region = self.region_map.get((objtype, target))
            if region is not None:
                return region
        return None


@dataclass
class LintIssue:
    line: int
    message: str
    fixed: bool = False


@dataclass
class LintWork:
    doctree: nodes.document
    content_lines: list[str]
    fix: bool
    fixed: bool


CHECKS = {}


def check(name: str):
    """Decorator to register a lint check function."""

    def decorator(func):
        CHECKS[name] = func
        return func

    return decorator


# pat/repl pairs for references styles.
REF_FIXES = [
    (
        # :class:`MyClass` -> :class:`!MyClass`
        r":{reftype}:`[~.]?{target}`",
        r":{reftype}:`!{target}`",
    ),
    (
        # :class:`Some Class <mymodule.MyClass>` -> :class:`!Some Class`
        r":{reftype}:`([^<]+?)\s*<{target}>`",
        r":{reftype}:`!\1`",
    ),
]

# Sphinx roles that are explicit references elsewhere and should always be links.
ALWAYS_REF = {"doc", "download", "numref", "ref"}


@check("self")
def check_self_links(work: LintWork) -> Iterable[LintIssue]:
    for ref in find_self_refs(work.doctree):
        line = node_line_number(ref)
        reftype = ref.get("reftype")  # type: ignore
        target = ref.get("reftarget")  # type: ignore
        fixed = False
        if work.fix:
            for pat, repl in REF_FIXES:
                fixed = resub_in_rst_line(
                    lines=work.content_lines,
                    line_num=line - 1,
                    pat=pat.format(reftype=reftype, target=re.escape(target)),
                    repl=repl.format(reftype=reftype, target=target),
                    count=1,
                )
                if fixed:
                    break
            work.fixed |= fixed
            if not fixed:
                print(f"Line {line}: Couldn't fix self-link to :{reftype}:`{target}`")
                print(f"Line was: {work.content_lines[line - 1]!r}")
        yield LintIssue(line, f"self-link to :{reftype}:`{target}`", fixed=fixed)


class RefFinder(nodes.SparseNodeVisitor):
    """Visitor for nodes to track class context and find self-references."""

    def __init__(self, doctree: nodes.document) -> None:
        super().__init__(doctree)
        self.resolver = Resolver(doctree)
        self.self_refs: list[nodes.Node] = []
        self.class_stack: list[str] = []
        self.pushed_class: list[bool] = []

    def visit_pending_xref(self, node: addnodes.pending_xref) -> None:
        line = node_line_number(node)
        reftype = node.get("reftype")
        if reftype in ALWAYS_REF:
            # :ref: references are explicitly meant to be links, so should
            # never be unlinked. I don't know why someone would put a :ref:
            # to its own section, but who are we to judge?
            return
        target = node.get("reftarget")

        if reftype == "meth" and "." not in target:
            if self.class_stack:
                target = f"{self.class_stack[-1]}.{target}"
            else:
                # Method reference without class context: can't resolve, so skip.
                target = None

        if target:
            region = self.resolver.find_region(reftype, target)
            if region is not None and region.start <= line <= region.end_total:
                self.self_refs.append(node)

    def visit_desc(self, node: addnodes.desc) -> None:
        objtype = node.get("objtype")
        pushed = False
        if objtype == "class":
            self.class_stack.append(node.children[0].get("fullname"))
            pushed = True
        elif objtype == "method" and not self.class_stack:
            self.class_stack.append(node.children[0].get("class"))
            pushed = True
        self.pushed_class.append(pushed)

    def depart_desc(self, node: addnodes.desc) -> None:
        if self.pushed_class.pop():
            self.class_stack.pop()

    # Have to explicitly pass on unknown nodes, and Python docs have a bunch.
    def unknown_visit(self, node):  # pragma: no cover
        pass

    def unknown_departure(self, node):  # pragma: no cover
        pass


def find_self_refs(doctree: nodes.document) -> Iterable[nodes.Node]:
    """Find references that point to the same region they are in."""

    finder = RefFinder(doctree)
    doctree.walkabout(finder)
    return finder.self_refs


@check("paradup")
def check_duplicate_refs_in_paragraph(work: LintWork) -> Iterable[LintIssue]:
    """Check references that appear more than once in the same paragraph."""
    if work.fix:
        raise Exception("Fixing is not available for --check=paradup")

    for ref in find_duplicate_refs(work.doctree):
        line = node_line_number(ref)
        reftype = ref.get("reftype")  # type: ignore
        target = ref.get("reftarget")  # type: ignore
        yield LintIssue(line, f"duplicate :{reftype}:`{target}` in paragraph")


def find_duplicate_refs(doctree: nodes.document) -> Iterable[nodes.Node]:
    """Find references that appear more than once in the same paragraph."""
    for para in doctree.findall(nodes.paragraph):
        refs_by_target = defaultdict(list)
        for ref in para.findall(addnodes.pending_xref):
            reftype = ref.get("reftype")
            if reftype in ALWAYS_REF:
                # :ref: references are explicitly meant to be links, so should
                # never be unlinked.
                continue
            target = ref.get("reftarget")
            assert reftype and target, (
                f"Reference missing reftype or target: {ref}\n{node_traceback(ref)}"
            )
            refs_by_target[(reftype, target)].append(ref)

        for refs in refs_by_target.values():
            if len(refs) > 1:
                yield from refs[1:]


@dataclass
class LintResult:
    content: str
    issues: list[LintIssue]
    fixed: bool


def lint_content(content: str, fix: bool, checks: set[str]) -> LintResult:
    doctree = parse_rst(content)
    work = LintWork(
        content_lines=content.splitlines(keepends=True),
        doctree=doctree,
        fix=fix,
        fixed=False,
    )

    issues = []
    for check_name in checks:
        issues.extend(CHECKS[check_name](work))

    result = LintResult(
        content="".join(work.content_lines),
        issues=issues,
        fixed=work.fixed,
    )

    return result
